#!/usr/bin/perl -w
#
# smartupdate.pl - updates mame roms based on their checksum using www.mame.dk
#
# Copyright (C) 2001 Zachary P. Landau <kapheine@hypa.net>
#
# version 0.5.3
# usage: smartupdate.pl [-a] [-u] [-m #] [-t #] [-r dir] [-d 0-5] [-q] [-v]
#
# -a: auto-blacklisting
# -u: update instead of just pretending
# -m: number of updates in a row allowed before autoblacklisted
# -t: timeout when accessing pages
# -r: directory that contains roms
# -d: debugging level 0-5
# -q: quiet
# -v: version display
#
# dont forget the required program mamedkget, available at:
# http://kapheine.hypa.net/smartupdate/mamedkget.txt
#

require LWP::UserAgent;
use POSIX;
use File::Basename;
use Debug;
use strict;

my $version = '0.5.3';
###################################################################################
# User Configuration
$Debug::level = 0;							# 0-5
my $quiet = 0;
my $update = 0;								# 0, just displays out of date roms
											# 1, will update them
my $autoblacklist = 1;						# automatically add roms to blacklist
my $max_updates = 2;						# if it wants to update for the third
											# time in a row, blacklist
my $update_script = 'mamedkget';			# should accept rom name, if it
											# accepts a filename, change update_rom
my $rom_dir = '/usr/local/games/xmame/roms';	# exclude trailing /
my $zip = '/usr/bin/unzip -vl';				# i will try to make this more portable
my $useragent = "smartupdate $version";
my $timeout = 30;
my $checklist_file = "$ENV{'HOME'}/.smamecl";
my $blacklist_file = "$ENV{'HOME'}/.smamebl";
# End User Configuration
###################################################################################

###################################################################################
# System Vars
my $ua;
my $request;
my $response;
my $result;
my $rom_name = 'unknown';
my $rom_file;
my $unused;
my $suffix;
my $entry;
my @blacklist;
my %checklist;
# End System Vars
###################################################################################

sub startup {
	debug(3, "setting up useragent.\n");
	$ua = LWP::UserAgent->new();
	$ua->timeout($timeout);
	$ua->agent($useragent);
}

sub shutdw {
	debug(3, "Running shutdown code\n");
	close(MAMEROMS);

	&write_blacklist;
	&write_checklist;
}

sub help {

	die "smartupdate $version\n\n",
		"\t-a, --autoblacklist\t\tauto blacklisting support\n",
		"\t-u, --update\t\t\tupdate instead of just pretending\n",
		"\t-m, --max-updates #\t\tnumber of updates in a row allowed before autoblacklisted\n",
		"\t-t, --timeout #\t\t\ttimeout when accessing pages\n",
		"\t-r, --romdir dir\t\tdirectory containing roms\n",
		"\t-d, --debug #\t\t\tdebugging level 0-5\n",
		"\t-q, --quiet\t\t\tno output\n",
		"\t-v, --version\t\t\tdisplay program version\n",
		"\n";

}

sub read_blacklist {

	debug(3, "opening blacklist at $blacklist_file for reading\n");

	if (!open(BLACKLIST, $blacklist_file)) {
		debug(4, 'blacklist file does not exist, ignoring\n');
		return;
	}
	push(@blacklist, chop($_)) while (<BLACKLIST>);
	close(BLACKLIST);
}

sub write_blacklist {
	my $bl;

	debug(3, "opening blacklist at $blacklist_file for writing\n");
	open(BLACKLIST, ">$blacklist_file");

	while ($bl = shift(@blacklist)) {
		debug(4, "BLACKLIST: $bl\n");
		print BLACKLIST $bl;
	}
	close(BLACKLIST);
}

sub read_checklist {
	my $rom;
	my $number;

	debug(3, "opening checklist at $checklist_file for reading\n");

	if (!open(CHECKLIST, $checklist_file)) {
		debug(4, 'checklist file does not exist, ignoring\n');
		return;
	}
	while (<CHECKLIST>) {
		($rom, $number) = split(/:/, $_);
		chop($number);
		$checklist{"$rom"} = $number;
	}

	close(CHECKLIST);
}

sub write_checklist {
	my $rom;
	my $number;

	debug(3, "opening checklist at $checklist_file for writing\n");

	open(CHECKLIST, ">$checklist_file");

	while (($rom, $number) = each (%checklist)) {
		print CHECKLIST "$rom:$number\n";
	}

	close(CHECKLIST);
}


# check_rom($romname)
sub check_rom {
	my ($rom) = @_;
	my $rom_url;
	
	debug(4, "retrieving page for $rom\n");

	$rom_url = "http://www.mame.dk/gameinfo/$rom";

	$request = HTTP::Request->new('GET', $rom_url);

	if (!$request) {
		die("connection to $rom_url failed\n");
	}

	$response = $ua->request($request);

	if (!$response) {
		die("connection to $rom_url failed [2]\n");
	}

	if (!$response->content) {
		die("url $rom_url returned no content\n");
	}

	&rom_checksum($response->content, $rom);
}

# rom_checksum($html_page, $romname)
sub rom_checksum {
	my ($html_page, $romname) = @_;
	my @html_lines;
	my @zip_output;
	my $i;
	my %checksum_r;											# prom->checksum (mame.dk reported)
	my %checksum_l;											# prom->checksum (local file reported)
	my $checksum;											# temp var

	@html_lines = split(/\n/, $html_page);

	# get checksums from mame.dk
	for ($i = 0; $i < @html_lines; $i++) {
		if ($html_lines[$i] =~ m!(.+):<\/B><\/TD> <TD valign="top" class="stdtext">(\w+) \d+ In  <a href="/download/rom/(\w*)/">!) {
			debug(5, "romname: $3 prom: $1 ($2)\n");
			if ($3 eq "$romname") {							# sometimes there are some other roms it uses, ignore em
				$checksum_r{"$1"} = $2;						# they will be updated later if you have them
			}
		} 
	}

	# get checksums from local files
	# note: this should be made more portable, using Archive::Zip probably
	@zip_output = `$zip ${rom_dir}/${romname}.zip`;

	while ($_ = shift(@zip_output)) {
		if (/\d+ .+ \d+  \d\d%  \d\d-\d\d-\d\d \d\d:\d\d  (\w+)  (.+)/) {
			debug(5, "prom $2 ($1)\n");
			$checksum_l{"$2"} = $1;
		}
	}

	# now lets look for differences
	# note about comparison below: sometimes actual file contains more than displayed on the page and we
	# arent smart enough to figure out what requires an update. instead, only care if listing shows MORE
	# than local file
	if ((keys(%checksum_l)-1) > keys(%checksum_r)) {		# -1 because textfiles are included, but not reported on the page
		debug(4, "number of files not equal for $romname\n");
		update_rom($romname);
		return;
	}

	foreach $checksum (keys(%checksum_l)) {
		if (exists($checksum_r{"$checksum"})) {
			if ($checksum_r{"$checksum"} ne $checksum_l{"$checksum"}) {
				debug(4, "$checksum checksums not equal: " . $checksum_r{"$checksum"} . " v. " . $checksum_l{"$checksum"} . "\n");
				update_rom($romname);
				return;
			}
		}
	}
}

# update_rom($romname)
sub update_rom {
	my ($rom) = @_;
	my $output;

	$checklist{"$rom"}++;									# increase number of times downloaded

	if ($checklist{"$rom"} >= $max_updates) {
		print "BLACKLISTED: $rom\n";
		push(@blacklist, $rom);
		delete($checklist{"$rom"});							# don't bother to keep it in checklist, it's blacklisted
		return;
	}

	if ($Debug::level || !$quiet) {
		print "fetching $rom...";
	}

	if ($update == 1) {
		$output = `$update_script $rom $rom_dir/$rom.zip`;
		print "\n";

		# oh my, i need to add error codes to the fetching script
		if ($output =~ /Couldn't find final download location/) {
			print "WARNING: update script could not find $rom, skipping\n";
		}
	}

	if (!$update) {			# don't run it twice
			print "\n";
	}

}

$_ = join(' ', @ARGV);
if (/--autoblacklist|-a/) {
	$autoblacklist = 1;
}
if (/--update|-u/) {
	$update = 1;
}
if (/(--max-updates|-m) (\d)/) {
	$max_updates = $2;
}
if (/(--timeout|-t) (\d)/) {
	$timeout = $2;
}
if (/(--romdir|-r) (.+)/) {
	$rom_dir = $2;
}
if (/(--debug|-d) (\d)/) {
	$Debug::level = $2;
}
if (/--help|-h/) {
	&help;
}
if (/--quiet|-q/) {
	$quiet = 1;
}
if (/--version|-v/) {
	die("$0 version $version\n");
}

$SIG{'INT'} = 'shutdw';
&startup;
&read_blacklist;
&read_checklist;

if (!opendir(MAMEROMS, $rom_dir)) {
	die("could not access $rom_dir\n");
}

SKIP: while ($rom_file = readdir(MAMEROMS)) {
	$rom_file = lc($rom_file);
	if ($rom_file =~ m/\.zip$/) {
		($rom_name, $unused, $suffix) = fileparse($rom_file, '\.zip');
		foreach $entry (@blacklist) {
			if ($entry eq $rom_name) {
				debug(3, "skipping $rom_name, blacklisted\n");
				goto SKIP;								# muhahaha
			}
		}
		&check_rom($rom_name);
	}
}

&shutdw;
